{{#general_title}}
# {{{title}}}

{{/general_title}}
{{#versions}}

{{#sections}}
### {{{label}}}

{{#commits}}
* {{{subject}}} [{{{author}}}]
{{#body}}

{{{body_indented}}}
{{/body}}

{{/commits}}
{{/sections}}

{{/versions}}
